
(* Here is a basic, non generic, version of insertion sort. *)
fun insert (a, nil) = [a]
 |  insert (a, b::rest) = if a < b then a::b::rest else b::insert(a, rest);

fun sort nil = nil
 |  sort (a :: rest) = insert( a, sort rest);


(* Here is a generic version of insertion sort. *)
fun insert (a, nil, cmp) = [a]
 |  insert (a, b::rest, cmp) = if cmp(a,b) then a::b::rest else b::insert(a, rest, cmp);

fun sort (nil, cmp) = nil
 |  sort (a::rest, cmp) = insert(a, sort(rest, cmp), cmp);


(* This is the generic insertion sort with insert as a nested function. *)
fun sort (nil, cmp) = nil
 |  sort (a::rest, cmp) =
     let
       fun insert (a, nil) = [a]
        |  insert (a, b::rest) = if cmp(a,b) then a::b::rest else b::insert(a, rest);
     in
       insert(a, sort(rest, cmp))
     end;


(* Call sort using an anonymous function for the comparator. *)
sort( [(2,3), (3,4), (~5,1), (0,0), (6,3), (8,2)], fn ((a,b),(x,y)) => a+b<x+y );


(* This is a curried version of the generic insertion sort. *)
fun sort cmp nil = nil
 |  sort cmp (a::rest) =
     let
       fun insert (a, nil) = [a]
        |  insert (a, b::rest) = if cmp(a,b) then a::b::rest else b::insert(a, rest);
     in
       insert(a, sort cmp rest)
     end;


sort (fn ((a,b),(x,y)) => a+b<x+y) [(2,3), (3,4), (~5,1), (0,0), (6,3), (8,2)];


(* Make a new kind of sort function. *)
val sortPairs = sort (fn ((a,b),(x,y)) => a+b<x+y);

(* Use this new sort function. *)
sortPairs [(2,1), (5,6), (~3,5), (0,8)];
sortPairs [(1,9), (2,8), (3,7), (4,6), (5,5)];


(* Create another kind of sort function. *)
val sortAbsolute = sort (fn (a,b) => abs(a)<abs(b));

(* Use it. *)
sortAbsolute [2, ~4, 3, ~3, 5, 0, ~1, 8, ~10];

(* Notice how the curried sort function is kind of like
   a "factory" for creating new kinds of sort functions. *)